#version 330 core
layout(location = 0) out vec4 out_color;

uniform vec3 uniform_diffuse;
uniform vec3 uniform_specular;
uniform float uniform_shininess;
uniform float uniform_has_texture;
uniform sampler2D diffuse_texture;

// Camera Properties
uniform vec3 uniform_camera_position;

// Light Source Properties
uniform mat4 uniform_light_projection_matrix;
uniform mat4 uniform_light_view_matrix;
uniform vec3 uniform_light_position;
uniform vec3 uniform_light_direction;
uniform vec3 uniform_light_color;
uniform float uniform_light_umbra;
uniform float uniform_light_penumbra;
uniform float uniform_light_exponent;
uniform int uniform_cast_shadows;
uniform sampler2D shadowmap_texture;

float uniform_constant_bias = 0.001;

in vec2 f_texcoord;
in vec3 f_position_wcs;
in vec3 f_normal;

#define PI 3.14159

float shadow_nearest(vec3 light_space_xyz)
{
	
	
	// sample shadow map
	float shadow_map_z = texture2D(shadowmap_texture, light_space_xyz.xy).r;

	// + shaded -> 0.0 
	// - lit -> 1.0
	return clamp(-sign((light_space_xyz.z - uniform_constant_bias) - shadow_map_z), 0.0, 1.0);
}

float shadow_multi(vec3 light_space_xyz)
{
	// sample shadow map

	float vis = 0.0f;
	for (int i=-2;i<3;i++)
	for (int j=-2;j<3;j++)
	{
		float r = 0.0005;
		vec2 off = vec2(i*r,j*r);
		float shadow_map_z = texture2D(shadowmap_texture, light_space_xyz.xy+off).r;
		vis += clamp(-sign((light_space_xyz.z - uniform_constant_bias) - shadow_map_z), 0.0, 1.0);
	}
	
	return vis/25.0;
}

// 1 sample per pixel
float shadow(vec3 pwcs)
{
	vec4 plcs = uniform_light_projection_matrix * uniform_light_view_matrix * vec4(pwcs, 1.0);
	plcs /= plcs.w;
	plcs.xy = (plcs.xy + 1) * 0.5;
	
	// check that we are inside light clipping frustum
	//if (plcs.x < 0.0) return 0.0; if (plcs.y < 0.0) return 0.0; if (plcs.x > 1.0) return 0.0; if (plcs.y > 1.0) return 0.0;
	if ((clamp(plcs.xy, vec2(0,0), vec2(1,1)) - plcs.xy) != vec2(0,0)) return 0.0;

	// set scale of light space z vaule to [0, 1]
	plcs.z = 0.5 * plcs.z + 0.5;

	
	// sample shadow map
	//return shadow_nearest(plcs.xyz);
	return shadow_multi(plcs.xyz);
	
}

float compute_spotlight(vec3 vertex_to_light_direction)
{
	float light_cosine_umbra = cos(radians(0.5 * uniform_light_umbra));
	float light_cosine_penumbra = cos(radians(0.5 * uniform_light_penumbra));
	float spoteffect = 1;
	float angle_vertex_spot_dir = dot(normalize(-vertex_to_light_direction), uniform_light_direction);

	if (angle_vertex_spot_dir > light_cosine_umbra) 
	{
		spoteffect = 1;
	}
	else if(angle_vertex_spot_dir > light_cosine_penumbra) 
	{
		spoteffect = smoothstep(light_cosine_penumbra, light_cosine_umbra, angle_vertex_spot_dir);
	}
	else 
		spoteffect = 0;
	
	return spoteffect;
}


//-------------------------------------------------------------------------------------------------------------------------------------

// some global consts
const float pi=3.14159;
const float unit_scale = 1;
const int num_samples = 64;

// Beckmann Shadowing-Masking
float GeometricBeckmann(float HI, float m)
{
  //float tan_a = tan(acos(NI));
	//float a = 1.0 / (m * tan_a);
	float a = HI / (m * sqrt(1.0 - (HI * HI)));
	float a2 = a * a;
	return (a < 1.6) ? (3.535 * a + 2.181 * a2) / (1.0 + 2.276 * a + 2.577 * a2) : 1.0;
}

float DistributionBeckmann(float HN, float m)
{
	// Beckmann
	float HN2 = HN*HN;
	float m2 = max(float(0.001), m*m);
	float D = exp((HN2 - 1.0) / (m2*HN2 + 0.01)) / (pi*m2*HN2*HN2 + 0.01);

	return D;
}

// Modified Shlick's approximation 
// Pass the reflected color at normal angle directly
vec3 Fresnel(vec3 R0, float HV)
{
  vec3 R=R0*R0;
  
  float inc = 1.0 - HV;
  float inc2=inc*inc;
  inc2=inc2*inc2*inc;
   return min(vec3(1.0), R + (vec3(1,1,1)-R)*inc2);
}

 vec3 GetEnvironment(vec3 R, float roughness)
 {
   
   vec3 color_1 = vec3(0.0, 0.0, 0.0);
   vec3 color_2 = vec3(0.8, 0.7, 0.6);
   
   vec3 color_3 = vec3(0.2, 0.3, 1.2);
   vec3 color_4 = vec3(0.2, 0.2, 0.2);

   float gloss=(1.0-roughness);
   float f = 0.5+0.5*cos(2.0*pi*R.y)*cos(2.0*pi*R.x*sin(pi*R.x));
   f = smoothstep(0.49, 0.51, f);
  
   
   vec3 color = mix(color_1, color_2, f)/(pi*(0.01+0.3/roughness));
   f = 0.5-0.5*R.y;
   color = color + mix(color_3, color_4, f)/(pi*(0.01+0.3/roughness));
   
   return color;
  
 }


vec3 BRDF( in vec3 in_vec, in vec3 out_vec, in vec3 normal, 
                   in vec3 color, in float roughness, in float reflectivity, in float metal, in float transparency)
{
	// The halfway vector
	vec3 H = normalize(in_vec + out_vec);

	// dot products used frequently
	float NV = max(0.0, dot(normal, in_vec));
	float HV = max(0.0, dot(H, in_vec));
	float HN = max(0.0, dot(H, normal));
	float NL = max(0.0, dot(in_vec, normal));

	// Reflectivity vs Transmission
	vec3 C0 = vec3(1.0, 1.0, 1.0)*(1.0 - metal) + color*metal;
	vec3 F = Fresnel(C0, HV);
	
	vec3 K_s = F*reflectivity;

	// Determine diffuse scattering. Metals should not contribute
	// to diffuse illumination and transmitted light is also not
	// back-scattered.
	vec3 K_d = (vec3(1.0) - K_s)*(1.0 - metal)*(1.0 - transparency);
  vec3 diffuse = K_d * color / pi;
	
  if (K_s == vec3(0.0))
		 return diffuse;

	float D = DistributionBeckmann(HN, roughness);
	float G = GeometricBeckmann(HV, roughness);

	vec3 specular = K_s * float(0.25)*D* G / (float(0.1) + NL * NV);
	
	vec3 final_color = vec3(0.0);
	final_color += diffuse;
	final_color += specular;
	return final_color;
}

vec3 SampleDirection(in vec3 dir, float roughness, float s, float t)
{

  vec3 u = cross(dir, vec3(0,0,-1));
  u = normalize(u);
  vec3 v = cross(dir,u);
  
  float a = (s+10.0*gl_FragCoord.z+7.0*gl_FragCoord.y+7.6*gl_FragCoord.x)*7.3*pi;
  float cosa = cos(a);
  float sina = sin(a);
  
  vec3 newvec = s*(cosa*u+sina*v) + 0.2*dir*(1.0-s);
  newvec = normalize(newvec);
  return normalize(dir*(1.0-roughness)+newvec*roughness);

}

vec3 SampleUniformDirection(in vec3 dir, float s, float t)
{

  vec3 u = cross(dir, vec3(0,0,-1));
  u = normalize(u);
  vec3 v = cross(dir,u);
  
  float xi = 0.999*fract(s+1.147460*gl_FragCoord.z+7.346560*gl_FragCoord.y+7.634646*gl_FragCoord.x);
  float xj = 0.999*fract(s+1.1234980*gl_FragCoord.y+7.5670*gl_FragCoord.x+7.64237*gl_FragCoord.z);
    
  float phi = 2*PI * xj;
  float sq = sqrt(1 - xi*xi);
  return sq*cos(phi) * u + sq*sin(phi)* v + xi*dir;
  

}

vec3 SampleEnvironment(vec3 pos, vec3 normal, float roughness, vec3 color, float reflectivity, float metal)
{
  
  vec3 env_color = vec3(0.0);
  for (int i=0; i<num_samples; i++)
  {
    // estimate the reflection vector 
    vec3 R = reflect(-pos,normal);
    vec3 dir = SampleDirection(R, roughness, float(i)/float(num_samples),0 );
    vec3 env =  GetEnvironment(dir, roughness);
    env_color += BRDF( dir, pos, normal, color, roughness, reflectivity, metal, 0) * env /(1-roughness);// * dot(dir,normal)/roughness;

	dir =SampleUniformDirection(normal, float(i)/float(num_samples),0 );
    env =  GetEnvironment(dir, roughness);
    env_color += 2*PI*BRDF( dir, pos, normal, color, roughness, reflectivity, metal, 0) *  env * dot(dir,normal);
  }
  return 0.5*env_color/float(num_samples);

   

}



// material properties

const vec3 color = vec3(0.9, 0.8, 0.7);  // material color 
const float metal = 0.0;
const float roughness = 0.6;             // values 0.0-1.0                   
const float reflectivity = 0.3;
const float power = 1500.0 /PI;

void main(void) 
{	
	vec3 normal = normalize(f_normal);
	
	vec2 uv = f_texcoord;
	uv.y = 1-uv.y;
	vec3 occ = texture2D(diffuse_texture,uv).rgb;


	// compute the direction to the light source
	vec3 vertex_to_light_direction = normalize(uniform_light_position - f_position_wcs.xyz);
	// compute the distance to the light source
	float dist = distance(uniform_light_position, f_position_wcs.xyz) + 0.000001; // add small offset to avoid division with zero
	
	// direction to the camera
	vec3 viewDirection = normalize(uniform_camera_position - f_position_wcs.xyz);	
	vec3 halfVector = normalize(viewDirection + vertex_to_light_direction);
	
	// check if we have shadows
	float shadow_value = (uniform_cast_shadows == 1)? shadow(f_position_wcs.xyz) : 1.0;
	// compute the spot effect
	float spotEffect = compute_spotlight(vertex_to_light_direction);
	// compute the incident radiance
	vec3 Li = power * shadow_value * spotEffect * uniform_light_color / (dist*dist);
	float costh = max(0.0, dot(vertex_to_light_direction, normal));

	vec3 Lo = BRDF( vertex_to_light_direction, viewDirection, normal, color, roughness, reflectivity, metal, 0.0 ) * Li * costh; 


	//Lo += occ *occ* SampleEnvironment(viewDirection, normal, roughness, color, reflectivity, metal);






	out_color = vec4(Lo, 1.0);

}

